from typing import Any, Dict, Iterator, Optional

from flytekit import Resources, Workflow
from flytekit.models.security import Secret
from flytekit.types.file import JSONLFile
from flytekit.types.iterator import JSON

from .task import (
    BatchEndpointTask,
    BatchResult,
    DownloadJSONFilesTask,
    OpenAIFileConfig,
    UploadJSONLFileTask,
)


def create_batch(
    name: str,
    openai_organization: str,
    secret: Secret,
    config: Optional[Dict[str, Any]] = None,
    is_json_iterator: bool = True,
    file_upload_mem: str = "700Mi",
    file_download_mem: str = "700Mi",
) -> Workflow:
    """
    Uploads JSON data to a JSONL file, creates a batch, waits for it to complete, and downloads the output/error JSON files.

    :param name: The suffix to be added to workflow and task names.
    :param openai_organization: Name of the OpenAI organization.
    :param secret: Secret comprising the OpenAI API key.
    :param config: Additional config for batch creation.
    :param is_json_iterator: Set to True if you're sending an iterator/generator; if a JSONL file, set to False.
    :param file_upload_mem: Memory to allocate to the upload file task.
    :param file_download_mem: Memory to allocate to the download file task.
    """
    wf = Workflow(name=f"openai-batch-{name.replace('.', '')}")

    if is_json_iterator:
        wf.add_workflow_input("jsonl_in", Iterator[JSON])
    else:
        wf.add_workflow_input("jsonl_in", JSONLFile)

    upload_jsonl_file_task_obj = UploadJSONLFileTask(
        name=f"openai-file-upload-{name.replace('.', '')}",
        task_config=OpenAIFileConfig(openai_organization=openai_organization, secret=secret),
    )
    if config is None:
        config = {}
    batch_endpoint_task_obj = BatchEndpointTask(
        name=f"openai-batch-{name.replace('.', '')}",
        openai_organization=openai_organization,
        config=config,
    )
    download_json_files_task_obj = DownloadJSONFilesTask(
        name=f"openai-download-files-{name.replace('.', '')}",
        task_config=OpenAIFileConfig(openai_organization=openai_organization, secret=secret),
    )

    node_1 = wf.add_entity(
        upload_jsonl_file_task_obj,
        jsonl_in=wf.inputs["jsonl_in"],
    )
    node_2 = wf.add_entity(
        batch_endpoint_task_obj,
        input_file_id=node_1.outputs["result"],
    )
    node_3 = wf.add_entity(
        download_json_files_task_obj,
        batch_endpoint_result=node_2.outputs["result"],
    )

    node_1.with_overrides(requests=Resources(mem=file_upload_mem), limits=Resources(mem=file_upload_mem))
    node_3.with_overrides(requests=Resources(mem=file_download_mem), limits=Resources(mem=file_download_mem))

    wf.add_workflow_output("batch_output", node_3.outputs["result"], BatchResult)

    return wf



"""
.. currentmodule:: flytekitplugins.openai

.. autosummary::
   :template: custom.rst
   :toctree: generated/

   BatchEndpointAgent
   BatchEndpointTask
   BatchResult
   DownloadJSONFilesTask
   UploadJSONLFileTask
   OpenAIFileConfig
   create_batch
   ChatGPTAgent
   ChatGPTTask
"""

from .batch.agent import BatchEndpointAgent
from .batch.task import BatchEndpointTask, BatchResult, DownloadJSONFilesTask, OpenAIFileConfig, UploadJSONLFileTask
from .batch.workflow import create_batch
from .chatgpt.agent import ChatGPTAgent
from .chatgpt.task import ChatGPTTask



from dataclasses import dataclass
from enum import Enum
from typing import Dict, Optional

import cloudpickle

from flytekit import FlyteContextManager, lazy_module
from flytekit.core.type_engine import TypeEngine
from flytekit.extend.backend.base_agent import (
    AgentRegistry,
    AsyncAgentBase,
    Resource,
    ResourceMeta,
)
from flytekit.extend.backend.utils import convert_to_flyte_phase, get_agent_secret
from flytekit.models.literals import LiteralMap
from flytekit.models.task import TaskTemplate

openai = lazy_module("openai")
OPENAI_API_KEY = "FLYTE_OPENAI_API_KEY"


class State(Enum):
    Running = ["in_progress", "finalizing", "validating"]
    Success = ["completed"]
    Failed = ["failed", "cancelled", "cancelling", "expired"]

    @classmethod
    def key_by_value(cls, value) -> str:
        for member in cls:
            if value in member.value:
                return member.name


@dataclass
class BatchEndpointMetadata(ResourceMeta):
    openai_org: str
    batch_id: str

    def encode(self) -> bytes:
        return cloudpickle.dumps(self)

    @classmethod
    def decode(cls, data: bytes) -> "BatchEndpointMetadata":
        return cloudpickle.loads(data)


class BatchEndpointAgent(AsyncAgentBase):
    name = "OpenAI Batch Endpoint Agent"

    def __init__(self):
        super().__init__(task_type_name="openai-batch", metadata_type=BatchEndpointMetadata)

    async def create(
        self,
        task_template: TaskTemplate,
        inputs: Optional[LiteralMap] = None,
        **kwargs,
    ) -> BatchEndpointMetadata:
        ctx = FlyteContextManager.current_context()
        input_values = TypeEngine.literal_map_to_kwargs(
            ctx,
            inputs,
            {"input_file_id": str},
        )
        custom = task_template.custom

        async_client = openai.AsyncOpenAI(
            organization=custom.get("openai_organization"),
            api_key=get_agent_secret(secret_key=OPENAI_API_KEY),
        )

        custom["config"].setdefault("completion_window", "24h")
        custom["config"].setdefault("endpoint", "/v1/chat/completions")

        result = await async_client.batches.create(
            **custom["config"],
            input_file_id=input_values["input_file_id"],
        )
        batch_id = result.id

        return BatchEndpointMetadata(batch_id=batch_id, openai_org=custom["openai_organization"])

    async def get(
        self,
        resource_meta: BatchEndpointMetadata,
        **kwargs,
    ) -> Resource:
        async_client = openai.AsyncOpenAI(
            organization=resource_meta.openai_org,
            api_key=get_agent_secret(secret_key=OPENAI_API_KEY),
        )

        retrieved_result = await async_client.batches.retrieve(resource_meta.batch_id)
        current_state = retrieved_result.status

        flyte_phase = convert_to_flyte_phase(State.key_by_value(current_state))

        message = None
        if current_state in State.Failed.value and retrieved_result.errors:
            data = retrieved_result.errors.data
            if data and data[0].message:
                message = data[0].message

        result = retrieved_result.to_dict()

        ctx = FlyteContextManager.current_context()
        outputs = LiteralMap(
            literals={"result": TypeEngine.to_literal(ctx, result, Dict, TypeEngine.to_literal_type(Dict))}
        )

        return Resource(phase=flyte_phase, outputs=outputs, message=message)

    async def delete(
        self,
        resource_meta: BatchEndpointMetadata,
        **kwargs,
    ):
        async_client = openai.AsyncOpenAI(
            organization=resource_meta.openai_org,
            api_key=get_agent_secret(secret_key=OPENAI_API_KEY),
        )

        await async_client.batches.cancel(resource_meta.batch_id)


AgentRegistry.register(BatchEndpointAgent())



import asyncio
import logging
from typing import Optional

from flyteidl.core.execution_pb2 import TaskExecution

from flytekit import FlyteContextManager, lazy_module
from flytekit.core.type_engine import TypeEngine
from flytekit.extend.backend.base_agent import AgentRegistry, Resource, SyncAgentBase
from flytekit.extend.backend.utils import get_agent_secret
from flytekit.models.literals import LiteralMap
from flytekit.models.task import TaskTemplate

openai = lazy_module("openai")

TIMEOUT_SECONDS = 10
OPENAI_API_KEY = "FLYTE_OPENAI_API_KEY"


class ChatGPTAgent(SyncAgentBase):
    name = "ChatGPT Agent"

    def __init__(self):
        super().__init__(task_type_name="chatgpt")

    async def do(
        self,
        task_template: TaskTemplate,
        inputs: Optional[LiteralMap] = None,
        **kwargs,
    ) -> Resource:
        ctx = FlyteContextManager.current_context()
        input_python_value = TypeEngine.literal_map_to_kwargs(ctx, inputs, {"message": str})
        message = input_python_value["message"]

        custom = task_template.custom
        custom["chatgpt_config"]["messages"] = [{"role": "user", "content": message}]
        client = openai.AsyncOpenAI(
            organization=custom["openai_organization"],
            api_key=get_agent_secret(secret_key=OPENAI_API_KEY),
        )

        logger = logging.getLogger("httpx")
        logger.setLevel(logging.WARNING)

        completion = await asyncio.wait_for(client.chat.completions.create(**custom["chatgpt_config"]), TIMEOUT_SECONDS)
        message = completion.choices[0].message.content
        outputs = {"o0": message}

        return Resource(phase=TaskExecution.SUCCEEDED, outputs=outputs)


AgentRegistry.register(ChatGPTAgent())



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional

from mashumaro.mixins.json import DataClassJSONMixin

import flytekit
from flytekit import Resources, kwtypes, lazy_module
from flytekit.configuration import SerializationSettings
from flytekit.configuration.default_images import DefaultImages, PythonVersion
from flytekit.core.base_task import PythonTask
from flytekit.core.interface import Interface
from flytekit.core.python_customized_container_task import PythonCustomizedContainerTask
from flytekit.core.shim_task import ShimTaskExecutor
from flytekit.extend.backend.base_agent import AsyncAgentExecutorMixin
from flytekit.models.security import Secret
from flytekit.models.task import TaskTemplate
from flytekit.types.file import JSONLFile

openai = lazy_module("openai")


@dataclass
class BatchResult(DataClassJSONMixin):
    output_file: Optional[JSONLFile] = None
    error_file: Optional[JSONLFile] = None


class BatchEndpointTask(AsyncAgentExecutorMixin, PythonTask):
    _TASK_TYPE = "openai-batch"

    def __init__(
        self,
        name: str,
        openai_organization: str,
        config: Dict[str, Any],
        **kwargs,
    ):
        super().__init__(
            name=name,
            task_type=self._TASK_TYPE,
            interface=Interface(
                inputs=kwtypes(input_file_id=str),
                outputs=kwtypes(result=Dict),
            ),
            **kwargs,
        )

        self._openai_organization = openai_organization
        self._config = config

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {
            "openai_organization": self._openai_organization,
            "config": self._config,
        }


class OpenAIFileDefaultImages(DefaultImages):
    """Default images for the openai batch plugin."""

    _DEFAULT_IMAGE_PREFIXES = {
        PythonVersion.PYTHON_3_8: "cr.flyte.org/flyteorg/flytekit:py3.8-openai-batch-",
        PythonVersion.PYTHON_3_9: "cr.flyte.org/flyteorg/flytekit:py3.9-openai-batch-",
        PythonVersion.PYTHON_3_10: "cr.flyte.org/flyteorg/flytekit:py3.10-openai-batch-",
        PythonVersion.PYTHON_3_11: "cr.flyte.org/flyteorg/flytekit:py3.11-openai-batch-",
        PythonVersion.PYTHON_3_12: "cr.flyte.org/flyteorg/flytekit:py3.12-openai-batch-",
    }


@dataclass
class OpenAIFileConfig:
    openai_organization: str
    secret: Secret

    def _secret_to_dict(self) -> Dict[str, Optional[str]]:
        return {
            "group": self.secret.group,
            "key": self.secret.key,
            "group_version": self.secret.group_version,
            "mount_requirement": self.secret.mount_requirement.value,
        }


class UploadJSONLFileTask(PythonCustomizedContainerTask[OpenAIFileConfig]):
    _UPLOAD_JSONL_FILE_TASK_TYPE = "openai-batch-upload-file"

    def __init__(
        self,
        name: str,
        task_config: OpenAIFileConfig,
        container_image: str = OpenAIFileDefaultImages.find_image_for(),
        **kwargs,
    ):
        super().__init__(
            name=name,
            task_config=task_config,
            task_type=self._UPLOAD_JSONL_FILE_TASK_TYPE,
            executor_type=UploadJSONLFileExecutor,
            container_image=container_image,
            requests=Resources(mem="700Mi"),
            interface=Interface(
                inputs=kwtypes(
                    jsonl_in=JSONLFile,
                ),
                outputs=kwtypes(result=str),
            ),
            secret_requests=[task_config.secret],
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {
            "openai_organization": self.task_config.openai_organization,
            "secret_arg": self.task_config._secret_to_dict(),
        }


class UploadJSONLFileExecutor(ShimTaskExecutor[UploadJSONLFileTask]):
    def execute_from_model(self, tt: TaskTemplate, **kwargs) -> Any:
        secret = tt.custom["secret_arg"]
        client = openai.OpenAI(
            organization=tt.custom["openai_organization"],
            api_key=flytekit.current_context().secrets.get(
                group=secret["group"],
                key=secret["key"],
                group_version=secret["group_version"],
            ),
        )

        local_jsonl_file = kwargs["jsonl_in"].download()
        uploaded_file_obj = client.files.create(file=open(local_jsonl_file, "rb"), purpose="batch")
        return uploaded_file_obj.id


class DownloadJSONFilesTask(PythonCustomizedContainerTask[OpenAIFileConfig]):
    _DOWNLOAD_JSON_FILES_TASK_TYPE = "openai-batch-download-files"

    def __init__(
        self,
        name: str,
        task_config: OpenAIFileConfig,
        container_image: str = OpenAIFileDefaultImages.find_image_for(),
        **kwargs,
    ):
        super().__init__(
            name=name,
            task_config=task_config,
            task_type=self._DOWNLOAD_JSON_FILES_TASK_TYPE,
            executor_type=DownloadJSONFilesExecutor,
            container_image=container_image,
            requests=Resources(mem="700Mi"),
            interface=Interface(
                inputs=kwtypes(batch_endpoint_result=Dict),
                outputs=kwtypes(result=BatchResult),
            ),
            secret_requests=[task_config.secret],
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {
            "openai_organization": self.task_config.openai_organization,
            "secret_arg": self.task_config._secret_to_dict(),
        }


class DownloadJSONFilesExecutor(ShimTaskExecutor[DownloadJSONFilesTask]):
    def execute_from_model(self, tt: TaskTemplate, **kwargs) -> Any:
        secret = tt.custom["secret_arg"]
        client = openai.OpenAI(
            organization=tt.custom["openai_organization"],
            api_key=flytekit.current_context().secrets.get(
                group=secret["group"],
                key=secret["key"],
                group_version=secret["group_version"],
            ),
        )

        batch_result = BatchResult()
        working_dir = flytekit.current_context().working_directory

        for file_name, file_id in zip(
            ("output_file", "error_file"),
            (
                kwargs["batch_endpoint_result"]["output_file_id"],
                kwargs["batch_endpoint_result"]["error_file_id"],
            ),
        ):
            if file_id:
                file_path = str(Path(working_dir, file_name).with_suffix(".jsonl"))

                with client.files.with_streaming_response.content(file_id) as response:
                    response.stream_to_file(file_path)

                setattr(batch_result, file_name, JSONLFile(file_path))

        return batch_result



from typing import Any, Dict

from flytekit.configuration import SerializationSettings
from flytekit.core.base_task import PythonTask
from flytekit.core.interface import Interface
from flytekit.extend.backend.base_agent import SyncAgentExecutorMixin


class ChatGPTTask(SyncAgentExecutorMixin, PythonTask):
    """
    This is the simplest form of a ChatGPT Task, you can define the model and the input you want.
    """

    _TASK_TYPE = "chatgpt"

    def __init__(self, name: str, openai_organization: str, chatgpt_config: Dict[str, Any], **kwargs):
        """
        Args:
            name: Name of this task, should be unique in the project
            openai_organization: OpenAI Organization. String can be found here. https://platform.openai.com/docs/api-reference/organization-optional
            chatgpt_config: ChatGPT job configuration. Config structure can be found here. https://platform.openai.com/docs/api-reference/completions/create
        """

        if "model" not in chatgpt_config:
            raise ValueError("The 'model' configuration variable is required in chatgpt_config")

        task_config = {"openai_organization": openai_organization, "chatgpt_config": chatgpt_config}

        inputs = {"message": str}
        outputs = {"o0": str}

        super().__init__(
            task_type=self._TASK_TYPE,
            name=name,
            task_config=task_config,
            interface=Interface(inputs=inputs, outputs=outputs),
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Dict[str, Any]:
        return {
            "openai_organization": self.task_config["openai_organization"],
            "chatgpt_config": self.task_config["chatgpt_config"],
        }
